//
//  Application.h
//  ImageSegmentation
//
//  Created by kumada on 2013/03/18.
//  Copyright (c) 2013年 kumada. All rights reserved.
//

#ifndef __ImageSegmentation__Application__
#define __ImageSegmentation__Application__

#include <string>
#include <opencv2/imgproc/imgproc.hpp>
#include <vector>
#include <boost/filesystem.hpp>

class ImageSegmenter;

/*!
        This class provides the following functions:
            1) functions to create input data to the image segmentation
            2) functions to display the results given by the image segmentation
 */
class Application
{
        friend class ApplicationTester;
        
        //! This structure represents a pixel on an image
        struct Pixel
        {
                cv::Point point_;
                cv::Vec3b color_;
        };
        typedef std::vector<Pixel> Pixels;
        
public:
        Application();
        ~Application() = default;
        
        //! set an input image
        void set_src_image(const cv::Mat* src_image);
        
        //! show the input image using OpenCV
        void show_src_image();
        
        //! associate a mouse callback with the input image
        void set_mouse_callback_to_src_image();
        
        //! show an arbitrary image using OpenCV
        void show(const std::string& name, const cv::Mat& image);

        //! execute the main loop
        void loop();
        
        //! add the pixel that is specified by the user's mouse move
        void add_training_data(int key, const cv::Point& point, const cv::Vec3b& color);
        
        //! draw the pixels that are extracted by the user
        void draw_training_samples();
       
        //! permit only numeric keys
        /*!
                The user can specify numeric numbers from 0 to 9.
                These numbers distinguish among training data.
         */
        bool is_numeric_key(int flags) const
        {
                return ZeroKey_ <= flags && flags <= NineKey_;
        }
        
        const cv::Mat* get_src_image() const
        {
                return src_image_;
        }
        
        const std::string& get_src_name() const
        {
                return src_name_;
        }
        
        void set_image_segmenter(std::shared_ptr<ImageSegmenter> image_segmenter)
        {
                image_segmenter_ = image_segmenter;
        }
        
        void set_output_dir_path(const std::string& output_dir);
        
        int get_key() const
        {
                return key_;
        }
        
        void enable_drawing(bool flag)
        {
                draws_ = flag;
        }
        
        bool draws() const
        {
                return draws_;
        }
        
        void clear_windows();
        
private:
        Application(const Application&) = delete;
        Application& operator=(const Application&) = delete;
        Application(Application&&) = delete;
        Application& operator=(Application&&) = delete;

        //! properties used for drawing training data
        static constexpr int Radius_    {2};
        static constexpr int Thickness_ {2};
        
        //! properties used for the mouse move
        static constexpr int ZeroKey_   {48};
        static constexpr int NineKey_   {57};
        
        //! names used for cv::namedWindow
        std::vector<std::string>        window_names_;
        
        const cv::Mat*                  src_image_;
        const std::string               src_name_;
        boost::filesystem::path         output_dir_;
        
        //! image with training data specified by a user
        cv::Mat                         src_image_with_training_data_;
        
        std::shared_ptr<ImageSegmenter> image_segmenter_;
        
        //! training data the user selects
        std::map<int, Pixels>           training_samples_;
        
        //! input data for the image segementer
        std::vector<cv::Mat>            training_data_;
        
        int                             key_;
        bool                            draws_;
        
        void draw_training_sample(const Pixels& pixels, const cv::Scalar& color, cv::Mat& image) const;
        void create_training_data();
        void convert(const Pixels& pixels, cv::Mat& mat) const;
};

#endif /* defined(__ImageSegmentation__Application__) */
