//
//  ImageSegmenter.cpp
//  ImageSegmentation
//
//  Created by kumada on 2013/03/16.
//  Copyright (c) 2013年 kumada. All rights reserved.
//

#include "ImageSegmenter.h"
#include "Parameters.h"
#include "GaussianMixture.h"
#include "Log.h"

ImageSegmenter::ImageSegmenter(const Parameters& params)
        : parameters_(params)
{
        
}

void ImageSegmenter::set_src_image(const cv::Mat* src_image)
{
        src_image_ = src_image;
}

namespace
{
        bool check_model(const std::string& name, const cv::EM& model, const Parameters& params)
        {
                Log::println("\tmeans for " + name);
                const cv::Mat& means = model.get<cv::Mat>("means");
                assert(means.rows == params.num_gaussians_);
                assert(means.cols == ImageSegmenter::Dimension_);
                assert(means.type() == CV_64FC1);
                for ( int i = 0; i < params.num_gaussians_; ++i ) {
                        const double* ptr = means.ptr<double>(0);
                        Log::println("\t(" + std::to_string(ptr[0]) + ", " + std::to_string(ptr[1]) + ", " + std::to_string(ptr[2]) + ")");
                }
                return true;
        }
        
        bool check_training_data(const std::vector<cv::Mat>& data)
        {
                for ( const auto& d : data ) {
                        assert(d.type() == CV_64FC1);
                        assert(d.cols == ImageSegmenter::Dimension_);
                }
                return true;
        }
}

// test ok
void ImageSegmenter::train(const std::vector<cv::Mat>& training_data)
{
        assert(check_training_data(training_data));
        models_.clear();
        gaussian_mixtures_.clear();
        
        const std::size_t data_size = training_data.size();
        
        models_.reserve(data_size);
        gaussian_mixtures_.reserve(data_size);
        
        for ( std::size_t i = 0; i < data_size; ++i ) {
                const cv::Mat& data = training_data[i];
                cv::EM model {parameters_.num_gaussians_};
                model.train(data);
                assert(check_model(std::to_string(i) + "-label", model, parameters_));
                models_.push_back(std::move(model));
                gaussian_mixtures_.push_back(std::make_shared<GaussianMixture>(models_.back()));
        }
}

namespace
{
        const double calculate_probability(
                const cv::Mat&                         x_n,
                const std::shared_ptr<GaussianMixture> gmm
        ) {
                cv::Mat probs;
                gmm->calculate_probability(x_n, probs);
                assert(probs.type() == CV_64FC1);
                const int nclusters = gmm->get_nclusters();
                assert(probs.rows == 1 && probs.cols == nclusters);
                double prob = 0.0;
                for( int k = 0; k < nclusters; ++k ) {
                        const double* ptr = reinterpret_cast<const double*>(probs.data);
                        prob += ptr[k];
                }
                return prob;
        }
        
        constexpr double MinimumValue = -std::numeric_limits<double>::max();

}

void ImageSegmenter::predict(const cv::Mat& image)
{
        assert(image.type() == CV_8UC3);
        max_labels_.create(image.size(), CV_8UC1);
        probabilities_.create(image.rows * image.cols, static_cast<int>(models_.size()), CV_64FC1);
        
        const int w = image.cols;
        const int h = image.rows;
        for ( int j = 0, jw = j * w; j < h; ++j, jw += w ) {
                const std::uint8_t* src = image.ptr<std::uint8_t>(j);
                std::uint8_t* dst = max_labels_.ptr<std::uint8_t>(j);
                for ( int i = 0, i3 = i * Dimension_; i < w; ++i, i3 += Dimension_ ) {
                        // a point x_n
                        double data[] = {
                                src[0 + i3] / 255.0, // b
                                src[1 + i3] / 255.0, // g
                                src[2 + i3] / 255.0, // r
                        };
                        cv::Mat x_n(1, ImageSegmenter::Dimension_, CV_64FC1, data);
                        int n = jw + i;
                        
                        double p = MinimumValue;
                        std::size_t max_label = 0;
                        for ( std::size_t i = 0, mixtures_size = gaussian_mixtures_.size(); i < mixtures_size; ++i ) {
                                double probability = calculate_probability(x_n, gaussian_mixtures_[i]);
                                probabilities_.at<double>(n, static_cast<int>(i)) = probability;
                                if ( p < probability ) {
                                        p = probability;
                                        max_label = i;
                                }
                        }
                        dst[i] = max_label;
                }
        }
}

void ImageSegmenter::create_max_labeled_images()
{
        max_labeled_images_.clear();
        const std::size_t n = models_.size();
        max_labeled_images_.resize(n);
        for ( std::size_t i = 0; i < n; ++i ) {
                create_masked_image(max_labels_, static_cast<int>(i), max_labeled_images_[i]);
        }
}

void ImageSegmenter::create_masked_image(const cv::Mat& labels, int index, cv::Mat& image) const
{
        cv::Mat mask = (labels == index);
        image = cv::Scalar(0); // clear
        src_image_->copyTo(image, mask);
}


#if(UNIT_TEST)
#define BOOST_TEST_DYN_LINK

#include <boost/test/unit_test.hpp>
#include <random>

namespace
{
        std::default_random_engine e {};
        std::uniform_int_distribution<double> d {0, 1};
}

class ImageSegmenterTester
{
public:
        static cv::Mat create_label(int rows, int cols)
        {
                cv::Mat label(rows, cols, CV_8UC1);
                
                int center_x = cols / 2;
                int center_y = rows / 2;
                int radius = 90;
                for ( int j = 0; j < rows; ++j ) {
                        int dy = j - center_y;
                        int dy2 = dy * dy;
                        std::uint8_t* ptr = label.ptr<std::uint8_t>(j);
                        for ( int i = 0; i < cols; ++i ) {
                                int dx = i - center_x;
                                int dx2 = dx * dx;
                                if ( dx2 + dy2 < radius * radius ) {
                                        ptr[i] = 0; // fore
                                } else {
                                        ptr[i] = 1; // back
                                }
                        }
                }
                label.reshape(1, rows * cols);
                return label;
        }
        
        static void test_train()
        {
                Parameters params {};
                params.num_gaussians_ = 5;
                ImageSegmenter segmenter {params};
                const int num = 3;
                std::vector<cv::Mat> training_data(num);
                
                for ( auto& m : training_data ) {
                        create_data(m);
                }
                
                segmenter.train(training_data);
                const auto& models = segmenter.models_;
                BOOST_CHECK(models.size() == num);
                for ( const auto& model : models ) {
                        BOOST_CHECK(model.get<int>("nclusters") == params.num_gaussians_);
                }
                
                const auto& mixtures = segmenter.gaussian_mixtures_;
                BOOST_CHECK(mixtures.size() == num);
                for ( const auto& mixture : mixtures ) {
                        BOOST_CHECK(mixture->get_nclusters() == params.num_gaussians_);
                }
                
        }
        
private:
        static void create_data(cv::Mat& mat)
        {
                mat.create(100, ImageSegmenter::Dimension_, CV_64FC1);
                for ( int i = 0; i < mat.rows; ++i ) {
                        double* ptr = mat.ptr<double>(i);
                        ptr[0] = d(e);
                        ptr[1] = d(e);
                        ptr[2] = d(e);
                }
        }
};

BOOST_AUTO_TEST_CASE(TEST_ImageSegmenter)
{
        std::cout << "ImageSegmenter\n";
        ImageSegmenterTester::test_train();
}
#endif // UNIT_TEST

